#include <stdio.h>
#include <fcntl.h>

char *malloc(), *realloc(), *strchr(), *strcpy();
void colsort(), exit(), free(), perror(), qsort(), usage();

int rflg, nflg;
char *argv0, *ifs=" \t", buf[BUFSIZ*16];

main(argc, argv)
char **argv;
{
	extern optind;
	extern char *optarg;
	int c, fi, fo;
	char *ofile=NULL, *tfile="/tmp/cs00000";
	FILE *fpi, *fpo=stdout;

	argv0=argv[0];
	while ((c=getopt(argc, argv, "f:no:r")) != EOF)
		switch (c) {
		case 'f':	/* field separator */
			ifs=optarg;
			break;
		case 'n':	/* numeric sorting */
			nflg++;
			break;
		case 'o':	/* output file */
			ofile=optarg;
			(void)sprintf(&tfile[strlen(tfile)-5], "%d", getpid());
			if ((fpo=fopen(tfile, "w")) == NULL) {
				perror(tfile);
				exit(1);
			}
			break;
		case 'r':	/* reverse sort */
			rflg++;
			break;
		default:
			usage();
		}

	if (optind == argc)
		colsort(stdin, fpo);
	else
		while (optind < argc) {
			if ((fpi=fopen(argv[optind], "r")) == NULL)
				perror(argv[optind]);
			else {
				colsort(fpi, fpo);
				(void)fclose(fpi);
			}
			optind++;
		}

	if (ofile) {
		(void)fclose(fpo);
		(void)unlink(ofile);
		if (link(tfile, ofile) == -1) {
			if ((fi=open(tfile, O_RDONLY)) == -1) {
				perror(tfile);
				(void)unlink(tfile);
				exit(1);
			}
			if ((fo=creat(ofile, 0666)) == -1) {
				perror(tfile);
				(void)unlink(tfile);
				exit(1);
			}
			while ((c=read(fi, buf, sizeof(buf))) > 0)
				if (write(fo, buf, (unsigned)c) != c) {
					perror(ofile);
					(void)unlink(tfile);
					exit(1);
				}
			if (c == -1) {
				perror(tfile);
				(void)unlink(tfile);
				exit(1);
			}
		}
		(void)unlink(tfile);
	}
	exit(0);
}

void
usage()
{
	(void)fprintf(stderr, "Usage: %s [-nr] [-f<sep>] [-o<output>] [<file>...]\n", argv0);
	(void)fprintf(stderr, "\t-n\tnumeric sort\n");
	(void)fprintf(stderr, "\t-r\treverse sort\n");
	(void)fprintf(stderr, "\t-f<sep>\tinput field separators (first character is output separator)\n");
	(void)fprintf(stderr, "\t-o<out>\toutput file\n");
	exit(1);
}

srtfnc(a, b)
char **a, **b;
{
	int cmp;

	if (nflg)
		cmp=atoi(*a)-atoi(*b);
	else
		cmp=strcmp(*a, *b);

	return rflg?-cmp:cmp;
}

#define INCR 32

char **cl;
unsigned clsz;

void
colsort(fpi, fpo)
FILE *fpi, *fpo;
{
	int c, state=0;
	unsigned nfs=0;
	char *p;

	while ((c=getc(fpi)) != EOF)
		switch (state) {
		case 0:	/* waiting for non-field separator */
			if (c == '\n') {
prout:
				if (nfs) {
					qsort((char *)cl, nfs, sizeof(char *), srtfnc);
					(void)fprintf(fpo, "%s", cl[0]);
					free(cl[0]);
					for (c=1 ; c < nfs ; c++) {
						(void)fprintf(fpo, "%c%s", *ifs, cl[c]);
						free(cl[c]);
					}
					nfs=0;
				}
				(void)fprintf(fpo, "\n");
				if (fpo == stdout)
					(void)fflush(fpo);
				break;
			}
			if (strchr(ifs, c) != NULL)
				break;
			p=buf;
			*p++ = c;
			state=1;
			break;
		case 1: /* input data */
			if (c != '\n' && strchr(ifs, c) == NULL) {
				*p++ = c;
				break;
			}
			if (nfs == clsz) {
				if (clsz == 0)
					cl=(char **)malloc(sizeof(char *)*(clsz=INCR));
				else
					cl=(char **)realloc((char *)cl, sizeof(char *)*(clsz += INCR));
				if (cl == NULL) {
					perror(argv0);
					exit(1);
				}
			}
			*p++ = '\0';
			if ((cl[nfs]=malloc((unsigned)(p-buf))) == NULL) {
				perror(argv0);
				exit(1);
			}
			(void)strcpy(cl[nfs], buf);
			nfs++;
			state=0;
			if (c == '\n')
				goto prout;
			break;
		}
}
